#include <cassert>

#include <cstdlib>
using std::atoi;
using std::atof;
using std::exit;

#include <iostream>
using std::cerr;
using std::endl;

#include <fstream>
using std::ofstream;

#include <algorithm>
using std::find;

#include "ising.h"

#include "moments.hpp"

typedef unsigned long int counter;

void bad(void)
{
	cerr << "MonteCarlo (single|double|triple|cluster) (ferro|rand) LatticeSize Temperature" << endl;
	exit(1);
}

void advance(char m, size_t multiplier)
{
	attempts = 0;
	if (m == 'c')
		while (attempts < multiplier*spin.size()/8)
			cluster_update();
	else 
	{
		if (m == 's')
			while (attempts < multiplier*spin.size()/8)
			{
				attempt_single_flip();
				update();
			}
		else if (m == 'd')
			while (attempts < multiplier*spin.size()/8)
			{
				attempt_double_flip();
				update();
			}
		if (m == 't')
			while (attempts < multiplier*spin.size()/8)
			{
				attempt_triple_flip();
				update();
			}
	}
}

int main(int argc, char* argv[])
{
	if (argc != 5) bad();
	
	const char mode = argv[1][0];
	const char valid_modes[] = "sdtc";
	if (find(valid_modes,valid_modes+5,mode) == valid_modes+5) bad();

	const char init = argv[2][0];
	if (init != 'f' and init != 'r') bad();

	const size_t L = atoi(argv[3]);
	const double T = atof(argv[4]);	
	initialize(L,T,(init == 'f'));

	const counter equilibration = 100;
	advance(mode,equilibration);
	
	const counter bin_size = 100;
	const counter autocorrelation = 3;
	moments mag;
	moments en;
	ofstream fout_mag("mag.bins.dat",std::ios::app);
	ofstream fout_en("en.bins.dat",std::ios::app);
	for (counter i = 0; i < 150; ++i)
	{
		mag.clear();
		en.clear();
		for (counter m = 0; m < bin_size; ++m)
		{
			advance(mode,autocorrelation);
			measure_magnetization(mag);
			measure_energy(en);
		}
		mag.print(fout_mag,bin_size);
		en.print(fout_en,bin_size);
		//fout_mag << L << "\t" << T << "\t" << mag.average(bin_size) << endl;
		//fout_en << L << "\t" << T << "\t" << en.average(bin_size) << endl;
	}
}