#include <cstddef>
using std::size_t;

#include <cassert>

#include <cstring>
using std::strlen;

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
using std::ostream;
using std::cerr;
using std::endl;
using std::fixed;

#include <fstream>
using std::ofstream;

#include <algorithm>
using std::find;

#include <sstream>
using std::stringstream;

#include <algorithm>
using std::min;

#include <cmath>
using std::log;
using std::sqrt;

#include "ising.h"

const double Tc = 2./log(1.+sqrt(2.));
const double threeTc = 3.0*Tc; 
const double threeTc_inv = 1.0/threeTc;

void glutPrint(float x, float y, const char* text, float r, float g, float b, float a) 
{ 
    if(!text || !strlen(text)) return; 
    bool blending = false; 
    if(glIsEnabled(GL_BLEND)) blending = true; 
    glEnable(GL_BLEND); 
    glColor4f(r,g,b,a); 
    glRasterPos2f(x,y); 
    while (*text) { 
        glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *text); 
        text++; 
    } 
    if(!blending) glDisable(GL_BLEND); 
}

char mode;
void display()
{
#ifdef USE_LINEAR_CHAIN
	const double d = 0.75/L;
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
	{
		const double x = 1.5*(i+0.5)/L - 0.75;
		glBegin(GL_POLYGON);
			if (spin[i] < 0)
				glColor3f(0.1,0.8,0.1);
			else
				glColor3f(0.1,0.1,0.8);
				glVertex2f(x-d,-d);
				glVertex2f(x-d,+d);
				glVertex2f(x+d,+d);
				glVertex2f(x+d,-d);
		glEnd();
		glFlush();
	}
#endif
#ifdef USE_SQUARE_LATTICE
	const double d = 0.75/L;
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
	{
		const double x = 1.5*(i+0.5)/L - 0.75;
		const double y = 1.5*(j+0.5)/L - 0.75;
		glBegin(GL_POLYGON);
			if (spin[i+j*L] < 0)
				glColor3f(0.1,0.8,0.1);
			else
				glColor3f(0.1,0.1,0.8);
				glVertex2f(x-d,y-d);
				glVertex2f(x-d,y+d);
				glVertex2f(x+d,y+d);
				glVertex2f(x+d,y-d);
		glEnd();
		glFlush();
	}
#endif
	glBegin(GL_POLYGON);
		glColor3f(0.9,0.05,0.05);
		const double height = 1.9*min(1.0,T*threeTc_inv)-0.95;
		glVertex2f(0.85,height);
		glVertex2f(0.95,height);
		glVertex2f(0.95,-0.95);
		glVertex2f(0.85,-0.95);
	glEnd();
	glBegin(GL_LINE_LOOP);
		glColor3f(1.0,1.0,1.0);
		glVertex2f(0.85,0.95);
		glVertex2f(0.95,0.95);
		glVertex2f(0.95,-0.95);
		glVertex2f(0.85,-0.95);
	glEnd();
#ifdef USE_SQUARE_LATTICE	
	const double h0 = 1.9/3.0-0.95;
	glBegin(GL_LINES);
		glColor3f(1.0,1.0,1.0);
		glVertex2f(0.85,h0);
		glVertex2f(0.95,h0);
	glEnd();
#endif	
	stringstream s;
	s.precision(3);
	s << "temperature T = " << fixed << T;
	glutPrint(-0.88f, -0.88f,s.str().c_str(), 1.0f, 1.0f, 1.0f, 0.5f);
	s.str("");
	s << "update mode: " << (mode == 's' ? "single" : ( mode == 'd' ? "double" : ( mode == 't' ? "triple" : ( mode == 'c' ? "cluster" : "error" ) ) ) );
	glutPrint(-0.1f, -0.88f,s.str().c_str(), 1.0f, 1.0f, 1.0f, 0.5f);
	glFlush();
	glutSwapBuffers();
}

#ifdef USE_LINEAR_CHAIN
void generate_postscript(ostream &fout)
{
	const double ease0 = 0.075;
	const double ease1 = 0.925;
	fout << "%!PS-Adobe-3.0 EPSF-3.0" << endl;
	const double h = 612.*3/(L+2);
    fout << "%%BoundingBox: 0 0 612 " << h << endl;
    const double s = 612.0/(L+2);
    fout << s << " " << s << " scale" << endl;
    fout << "1 1 translate" << endl;
	fout << "0 setlinewidth" << endl;
	fout << "0 setgray" << endl;

	for (size_t i = 0; i < L; ++i)
		if (spin[i] > 0)
		{		
			fout << "newpath" << endl;
				fout << i+ease0 << " " << ease0 << " moveto" << endl;
				fout << i+ease1 << " " << ease0 << " lineto" << endl;
				fout << i+ease1 << " " << ease1 << " lineto" << endl;
				fout << i+ease0 << " " << ease1 << " lineto" << endl;
				fout << "closepath" << endl;
				fout << "fill" << endl;		
		}
}
#endif

#ifdef USE_SQUARE_LATTICE
void generate_postscript(ostream &fout)
{
	const double ease0 = 0.075;
	const double ease1 = 0.925;
	fout << "%!PS-Adobe-3.0 EPSF-3.0" << endl;
    fout << "%%BoundingBox: 0 0 612 612" << endl;
    const double s = 612.0/(L+2);
    fout << s << " " << s << " scale" << endl;
    fout << "1 1 translate" << endl;
	fout << "0 setlinewidth" << endl;
	fout << "0 setgray" << endl;
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			if (spin[i+j*L] > 0)
			{		
				fout << "newpath" << endl;
				fout << i+ease0 << " " << j+ease0 << " moveto" << endl;
				fout << i+ease1 << " " << j+ease0 << " lineto" << endl;
				fout << i+ease1 << " " << j+ease1 << " lineto" << endl;
				fout << i+ease0 << " " << j+ease1 << " lineto" << endl;
				fout << "closepath" << endl;
				fout << "fill" << endl;		
			}
}
#endif

int startx, starty;
bool moving;
void mouse(int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON) 
	{
		if (state == GLUT_DOWN) 
		{
			moving = true;
			startx = x;
			starty = y;
		}
		if (state == GLUT_UP) 
		{
			moving = false;
		}
	}
}

unsigned int filenum = 0;
void keyboard(unsigned char key, int x, int y)
{
	if (key == 'p')
	{
		stringstream s;
		s << "config" << (++filenum) << ".ps";
		ofstream fout(s.str().c_str());
		generate_postscript(fout);
		fout.close();
	}
	else if (key == 's')
		mode = 's';
	else if (key == 'd')
		mode = 'd';
	else if (key == 't')
		mode = 't';
	else if (key == 'c')
		mode = 'c';
}

void motion(int x, int y)
{
	if (moving) 
		T += 0.011*GLfloat(starty - y);
	if (T > threeTc) T = threeTc;
	else if (T < 0.001) T = 0.001;
	startx = x;
	starty = y;		
}

const int delay = 0; // milliseconds
void animate(int)
{
	attempts = 0;
	if (mode == 'c')
		while (attempts < spin.size()/8)
			cluster_update();
	else 
	{
		if (mode == 's')
			while (attempts < spin.size()/8)
			{
				attempt_single_flip();
				update();
			}
		else if (mode == 'd')
			while (attempts < spin.size()/8)
			{
				attempt_double_flip();
				update();
			}
		if (mode == 't')
			while (attempts < spin.size()/8)
			{
				attempt_triple_flip();
				update();
			}
	}
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

void bad(void)
{
	cerr << "MonteCarlo_openGL (single|double|triple|cluster) (ferro|rand) LatticeSize" << endl;
	exit(1);
}

int main(int argc, char **argv)
{
	if (argc != 4) bad();
	mode = argv[1][0];
	const char valid_modes[] = "sdtc";
	if (find(valid_modes,valid_modes+5,mode) == valid_modes+5) bad();
	
	const char init = argv[2][0];
	if (init != 'f' and init != 'r') bad();
	
	const size_t L = atoi(argv[3]);
	initialize(L,1.0,(init == 'f'));
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("Ising model");
	glutDisplayFunc(display);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);	
	glutMainLoop();
}
